@@ -27,6 +27,9 @@
 
 #define MAX_NAME_FAILS 3
 
+#define MAX_SANE_ICON_SIZE 96
+#define FALLBACK_ICON_SIZE 24
+
 static gint unique_id = 0;
 
 enum
@@ -35,6 +38,7 @@ enum
     BUTTON_RELEASE,
     ACTIVATE,
     STATE_CHANGED,
+    SCROLL,
     LAST_SIGNAL
 };
 
@@ -45,6 +49,7 @@ enum
     PROP_0,
     PROP_PRIMARY_MENU,
     PROP_SECONDARY_MENU,
+    PROP_ICON_SIZE,
     N_PROPERTIES
 };
 
@@ -78,6 +83,7 @@ typedef struct
     gchar *tooltip_text;
     gchar *label;
     gboolean visible;
+    gint icon_size;
 
     guint owner_id;
     guint listener_id;
@@ -159,71 +165,93 @@ state_to_str (XAppStatusIconState state)
     }
 }
 
-static gint
-adjust_y_for_monitor_bounds (gint init_x,
-                             gint init_y,
-                             gint menu_height)
+static const gchar *
+direction_to_str (XAppScrollDirection direction)
 {
-    GdkDisplay *display = gdk_display_get_default ();
-    GdkMonitor *monitor;
-    GdkRectangle mrect;
-    gint bottom_edge_y;
-    gint ret_y;
-
-    ret_y = init_y;
-
-    monitor = gdk_display_get_monitor_at_point (display,
-                                                init_x,
-                                                init_y);
-
-    gdk_monitor_get_workarea (monitor, &mrect);
-
-    bottom_edge_y = mrect.y + mrect.height;
-
-    if ((init_y + menu_height) > bottom_edge_y)
+    switch (direction)
     {
-        ret_y = bottom_edge_y - menu_height;
+        case XAPP_SCROLL_UP:
+            return "Up";
+        case XAPP_SCROLL_DOWN:
+            return "Down";
+        case XAPP_SCROLL_LEFT:
+            return "Left";
+        case XAPP_SCROLL_RIGHT:
+            return "Right";
+        default:
+            return "Unknown";
     }
-
-    return ret_y;
 }
 
-typedef struct {
-    gint    x;
-    gint    y;
-    gint    position;
-    guint32 t;
-} PositionData;
-
-static void
-position_menu_cb (GtkMenu  *menu,
-                  gint     *x,
-                  gint     *y,
-                  gboolean *push_in,
-                  gpointer  user_data)
+static GdkEvent *
+synthesize_event (XAppStatusIcon *self,
+                  gint            x,
+                  gint            y,
+                  guint           button,
+                  guint           _time,
+                  gint            position,
+                  GdkWindow     **rect_window,
+                  GdkRectangle   *win_rect,
+                  GdkGravity     *rect_anchor,
+                  GdkGravity     *menu_anchor)
 {
-    GtkAllocation alloc;
-    PositionData *position_data = (PositionData *) user_data;
+  GdkDisplay *display;
+  GdkWindow *window;
+  GdkSeat *seat;
+  GdkWindowAttr attributes;
+  gint attributes_mask;
+  gint fx, fy;
 
-    *x = position_data->x;
-    *y = position_data->y;
+  display = gdk_display_get_default ();
+  seat = gdk_display_get_default_seat (display);
 
-    gtk_widget_get_allocation (GTK_WIDGET (menu), &alloc);
+  switch (position)
+  {
+      case GTK_POS_TOP:
+        fx = x;
+        fy = y - self->priv->icon_size;
+        *rect_anchor = GDK_GRAVITY_SOUTH_WEST;
+        *menu_anchor = GDK_GRAVITY_NORTH_WEST;
+        break;
+      case GTK_POS_BOTTOM:
+        fx = x;
+        fy = y;
+        *rect_anchor = GDK_GRAVITY_NORTH_WEST;
+        *menu_anchor = GDK_GRAVITY_SOUTH_WEST;
+        break;
+      case GTK_POS_LEFT:
+        fx = x - self->priv->icon_size;
+        fy = y;
+        *rect_anchor = GDK_GRAVITY_NORTH_EAST;
+        *menu_anchor = GDK_GRAVITY_NORTH_WEST;
+        break;
+      case GTK_POS_RIGHT:
+        fx = x;
+        fy = y;
+        *rect_anchor = GDK_GRAVITY_NORTH_WEST;
+        *menu_anchor = GDK_GRAVITY_NORTH_EAST;
+        break;
+  }
 
-    switch (position_data->position) {
-        case GTK_POS_BOTTOM:
-            *y = *y - alloc.height;
-            break;
-        case GTK_POS_RIGHT:
-            *x = *x - alloc.width;
-            *y = adjust_y_for_monitor_bounds (position_data->x, position_data->y, alloc.height);
-            break;
-        case GTK_POS_LEFT:
-            *y = adjust_y_for_monitor_bounds (position_data->x, position_data->y, alloc.height);
-            break;
-    }
+  attributes.window_type = GDK_WINDOW_CHILD;
+  win_rect->x = 0;
+  win_rect->y = 0;
+  win_rect->width = self->priv->icon_size;
+  win_rect->height = self->priv->icon_size;
+  attributes.x = fx;
+  attributes.y = fy;
+  attributes.width = self->priv->icon_size;
+  attributes.height = self->priv->icon_size;
+  attributes_mask = GDK_WA_X | GDK_WA_Y;
 
-    *push_in = TRUE;
+  window = gdk_window_new (NULL, &attributes, attributes_mask);
+  *rect_window = window;
+
+  GdkEvent *event = gdk_event_new (GDK_BUTTON_RELEASE);
+  event->any.window = window;
+  event->button.device = gdk_seat_get_pointer (seat);
+
+  return event;
 }
 
 static void
@@ -235,27 +263,31 @@ popup_menu (XAppStatusIcon *self,
             guint           _time,
             gint            panel_position)
 {
-    GdkDisplay *display;
-    GdkDevice *pointer;
+    GdkWindow *rect_window;
+    GdkEvent *event;
+    GdkRectangle win_rect;
+    GdkGravity rect_anchor, menu_anchor;
 
     g_debug ("XAppStatusIcon: Popup menu on behalf of application");
 
-    PositionData position_data = {
-        x, y, panel_position, _time
-    };
+    event = synthesize_event (self,
+                              x, y, button, _time, panel_position,
+                              &rect_window, &win_rect, &rect_anchor, &menu_anchor);
 
-    display = gdk_display_get_default ();
-    pointer = gdk_device_manager_get_client_pointer (gdk_display_get_device_manager (display));
+    g_object_set (G_OBJECT (menu),
+                  "anchor-hints", GDK_ANCHOR_SLIDE_X  | GDK_ANCHOR_SLIDE_Y  |
+                                  GDK_ANCHOR_RESIZE_X | GDK_ANCHOR_RESIZE_Y,
+                  NULL);
 
-    gtk_menu_popup_for_device (menu,
-                               pointer,
-                               NULL,
-                               NULL,
-                               position_menu_cb,
-                               &position_data,
-                               NULL,
-                               button,
-                               _time);
+    gtk_menu_popup_at_rect (menu,
+                            rect_window,
+                            &win_rect,
+                            rect_anchor,
+                            menu_anchor,
+                            event);
+
+    gdk_event_free (event);
+    gdk_window_destroy (rect_window);
 }
 
 static gboolean
@@ -285,48 +317,6 @@ should_send_activate (XAppStatusIcon *icon,
     return do_activate;
 }
 
-static gboolean
-appindicator_can_activate (XAppStatusIcon *icon)
-{
-    gpointer ptr;
-    gboolean has_activate = FALSE;
-
-    ptr = g_object_get_data (G_OBJECT (icon), "app-indicator-has-secondary-activate");
-
-    if (ptr && GPOINTER_TO_INT (ptr))
-    {
-        has_activate = TRUE;
-    }
-
-    return has_activate;
-}
-
-static gboolean
-handle_appindicator_button_press (XAppStatusIcon *icon,
-                                  guint           button,
-                                  guint           _time)
-{
-    if (g_object_get_data (G_OBJECT (icon), "app-indicator"))
-    {
-        if (button == GDK_BUTTON_MIDDLE || (button == GDK_BUTTON_PRIMARY && appindicator_can_activate (icon)))
-        {
-            g_debug ("XAppStatusIcon: sending activate for left- or middle-click event (libappindicator)");
-
-            g_signal_emit (icon, signals[ACTIVATE], 0,
-                           button,
-                           _time);
-        }
-        else
-        {
-            g_debug ("XAppStatusIcon: GtkStatusIcon (app-indicator) ignoring %s button press.", button_to_str (button));
-        }
-
-        return TRUE;
-    }
-
-    return FALSE;
-}
-
 static GtkWidget *
 get_menu_to_use (XAppStatusIcon *icon,
                  guint           button)
@@ -336,15 +326,6 @@ get_menu_to_use (XAppStatusIcon *icon,
     switch (button)
     {
         case GDK_BUTTON_PRIMARY:
-            if (g_object_get_data (G_OBJECT (icon), "app-indicator"))
-            {
-                if (!appindicator_can_activate (icon))
-                {
-                    menu_to_use = icon->priv->secondary_menu;
-                    break;
-                }
-            }
-
             menu_to_use = icon->priv->primary_menu;
             break;
         case GDK_BUTTON_SECONDARY:
@@ -374,15 +355,12 @@ handle_click_method (XAppStatusIconInterface *skeleton,
                  g_dbus_method_invocation_get_sender (invocation),
                  x, y, button_to_str (button), _time, panel_position_to_str (panel_position));
 
-        if (!handle_appindicator_button_press (icon, button, _time))
+        if (should_send_activate (icon, button))
         {
-            if (should_send_activate (icon, button))
-            {
-                g_debug ("XAppStatusIcon: native sending 'activate' for %s button", button_to_str (button));
-                g_signal_emit (icon, signals[ACTIVATE], 0,
-                               button,
-                               _time);
-            }
+            g_debug ("XAppStatusIcon: native sending 'activate' for %s button", button_to_str (button));
+            g_signal_emit (icon, signals[ACTIVATE], 0,
+                           button,
+                           _time);
         }
 
         icon->priv->have_button_press = TRUE;
@@ -434,20 +412,28 @@ handle_click_method (XAppStatusIconInterface *skeleton,
     return TRUE;
 }
 
-static void
-popup_gtk_status_icon_with_menu (XAppStatusIcon *icon,
-                                 GtkMenu        *menu,
-                                 GtkStatusIcon  *gtk_status_icon,
-                                 guint           button,
-                                 guint           activate_time)
+static gboolean
+handle_scroll_method (XAppStatusIconInterface *skeleton,
+                      GDBusMethodInvocation   *invocation,
+                      gint                     delta,
+                      XAppScrollDirection      direction,
+                      guint                    _time,
+                      XAppStatusIcon          *icon)
 {
-    gtk_menu_popup (menu,
-                    NULL,
-                    NULL,
-                    gtk_status_icon_position_menu,
-                    gtk_status_icon,
-                    button,
-                    activate_time);
+    g_debug ("XAppStatusIcon: received Scroll from monitor %s: "
+             "delta: %d , direction: %s , time: %u",
+             g_dbus_method_invocation_get_sender (invocation),
+             delta, direction_to_str (direction), _time);
+
+    g_signal_emit(icon, signals[SCROLL], 0,
+                  delta,
+                  direction,
+                  _time);
+
+    xapp_status_icon_interface_complete_scroll (skeleton,
+                                                invocation);
+
+    return TRUE;
 }
 
 static void
@@ -531,24 +517,17 @@ on_gtk_status_icon_button_press (GtkStatusIcon *status_icon,
     button = event->button.button;
     _time = event->button.time;
 
-    /* Button press equates to activate when there's no menu,
-     * but for appindicator, left click and right click will always
-     * bring up the same menu, and only a middle click should activate. */
-
     g_debug ("XAppStatusIcon: GtkStatusIcon button-press-event with %s button", button_to_str (button));
 
-    if (!handle_appindicator_button_press (icon, button, _time))
-    {
         /* We always send 'activate' for a button that has no corresponding menu,
          * and for middle clicks. */
-        if (should_send_activate (icon, button))
-        {
-            g_debug ("XAppStatusIcon: GtkStatusIcon activated by %s button", button_to_str (button));
+    if (should_send_activate (icon, button))
+    {
+        g_debug ("XAppStatusIcon: GtkStatusIcon activated by %s button", button_to_str (button));
 
-            g_signal_emit (icon, signals[ACTIVATE], 0,
-                           button,
-                           _time);
-        }
+        g_signal_emit (icon, signals[ACTIVATE], 0,
+                       button,
+                       _time);
     }
 
     calculate_gtk_status_icon_position_and_orientation (icon,
@@ -572,6 +551,7 @@ on_gtk_status_icon_button_release (GtkStatusIcon *status_icon,
                                    gpointer       user_data)
 {
     XAppStatusIcon *icon = user_data;
+    GtkWidget *menu_to_use;
     guint _time;
     guint button;
     gint x, y, orientation;
@@ -581,38 +561,10 @@ on_gtk_status_icon_button_release (GtkStatusIcon *status_icon,
 
     g_debug ("XAppStatusIcon: GtkStatusIcon button-release-event with %s button", button_to_str (button));
 
-    /* Icons originated in app-indicator have only a single menu, which displays on either
-     * right or left-click. */
-    if (g_object_get_data (G_OBJECT (status_icon), "app-indicator"))
-    {
-        if (button == GDK_BUTTON_PRIMARY || button == GDK_BUTTON_SECONDARY)
-        {
-            popup_gtk_status_icon_with_menu (icon,
-                                             GTK_MENU (icon->priv->primary_menu),
-                                             status_icon,
-                                             button,
-                                             _time);
-        }
-    }
-    else
-    {
-        /* Native icons can have two menus, so we must determine which to use based
-         * on the gtk icon event's button. */
-        GtkWidget *menu_to_use;
+    /* Native icons can have two menus, so we must determine which to use based
+     * on the gtk icon event's button. */
 
-        menu_to_use = get_menu_to_use (icon, button);
-
-        if (menu_to_use)
-        {
-            g_debug ("XAppStatusIcon: GtkStatusIcon popup menu for %s button", button_to_str (button));
-
-            popup_gtk_status_icon_with_menu (icon,
-                                             GTK_MENU (menu_to_use),
-                                             status_icon,
-                                             button,
-                                             _time);
-        }
-    }
+    menu_to_use = get_menu_to_use (icon, button);
 
     calculate_gtk_status_icon_position_and_orientation (icon,
                                                         status_icon,
@@ -620,6 +572,19 @@ on_gtk_status_icon_button_release (GtkStatusIcon *status_icon,
                                                         &y,
                                                         &orientation);
 
+    if (menu_to_use)
+    {
+        g_debug ("XAppStatusIcon: GtkStatusIcon popup menu for %s button", button_to_str (button));
+
+        popup_menu (icon,
+                    GTK_MENU (menu_to_use),
+                    x,
+                    y,
+                    button,
+                    _time,
+                    orientation);
+    }
+
     icon->priv->have_button_press = FALSE;
 
     g_signal_emit (icon, signals[BUTTON_RELEASE], 0,
@@ -720,7 +685,8 @@ typedef struct
 static SkeletonSignal skeleton_signals[] = {
     // signal name                                callback
     { "handle-button-press",                      handle_click_method },
-    { "handle-button-release",                    handle_click_method }
+    { "handle-button-release",                    handle_click_method },
+    { "handle-scroll",                            handle_scroll_method }
 };
 
 static gboolean
@@ -798,7 +764,6 @@ update_fallback_icon (XAppStatusIcon *self)
     }
 
     gtk_status_icon_set_tooltip_text (priv->gtk_status_icon, priv->tooltip_text);
-    gtk_status_icon_set_name (priv->gtk_status_icon, priv->name);
 
     if (priv->icon_name)
     {
@@ -1062,6 +1027,9 @@ xapp_status_icon_set_property (GObject    *object,
         case PROP_SECONDARY_MENU:
             xapp_status_icon_set_secondary_menu (XAPP_STATUS_ICON (object), g_value_get_object (value));
             break;
+        case PROP_ICON_SIZE:
+            XAPP_STATUS_ICON (object)->priv->icon_size = CLAMP (g_value_get_int (value), 0, MAX_SANE_ICON_SIZE);
+            break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
             break;
@@ -1084,6 +1052,9 @@ xapp_status_icon_get_property (GObject    *object,
         case PROP_SECONDARY_MENU:
             g_value_set_object (value, icon->priv->secondary_menu);
             break;
+        case PROP_ICON_SIZE:
+            g_value_set_int (value, icon->priv->icon_size);
+            break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
             break;
@@ -1097,6 +1068,7 @@ xapp_status_icon_init (XAppStatusIcon *self)
 
     self->priv->name = g_strdup_printf("%s", g_get_application_name());
     self->priv->state = XAPP_STATUS_ICON_STATE_NO_SUPPORT;
+    self->priv->icon_size = FALLBACK_ICON_SIZE;
 
     g_debug ("XAppStatusIcon: init: application name: '%s'", self->priv->name);
 
@@ -1146,9 +1118,13 @@ xapp_status_icon_dispose (GObject *object)
     g_clear_object (&self->priv->primary_menu);
     g_clear_object (&self->priv->secondary_menu);
 
-    g_signal_handlers_disconnect_by_func (self->priv->gtk_status_icon, on_gtk_status_icon_button_press, self);
-    g_signal_handlers_disconnect_by_func (self->priv->gtk_status_icon, on_gtk_status_icon_button_release, self);
-    g_clear_object (&self->priv->gtk_status_icon);
+    if (self->priv->gtk_status_icon != NULL)
+    {
+        g_signal_handlers_disconnect_by_func (self->priv->gtk_status_icon, on_gtk_status_icon_button_press, self);
+        g_signal_handlers_disconnect_by_func (self->priv->gtk_status_icon, on_gtk_status_icon_button_release, self);
+        g_object_unref (self->priv->gtk_status_icon);
+        self->priv->gtk_status_icon = NULL;
+    }
 
     tear_down_dbus (self);
 
@@ -1238,6 +1214,25 @@ xapp_status_icon_class_init (XAppStatusIconClass *klass)
                                                            GTK_TYPE_WIDGET,
                                                            G_PARAM_READWRITE));
 
+    /**
+     * XAppStatusIcon:icon-size:
+     *
+     * The icon size that is preferred by icon monitor/host - this is usually a product
+     * of some calculation based on the panel size.  It can be used by the client to size
+     * an icon to be saved as a file and its path sent to the host.
+     *
+     * If this value is 0 it has not been set, and its value can be unreliable if the host
+     * has multiple #XAppStatusIconMonitors active.
+     */
+    g_object_class_install_property (gobject_class, PROP_ICON_SIZE,
+                                     g_param_spec_int ("icon-size",
+                                                       "The icon size the monitor/host prefers",
+                                                       "The icon size that should be used, if the client is"
+                                                       " supplying absolute icon paths",
+                                                       0,
+                                                       96,
+                                                       0,
+                                                       G_PARAM_READWRITE));
 
     /**
      * XAppStatusIcon::button-press-event:
@@ -1312,6 +1307,25 @@ xapp_status_icon_class_init (XAppStatusIconClass *klass)
                       0,
                       NULL, NULL, NULL,
                       G_TYPE_NONE, 1, XAPP_TYPE_STATUS_ICON_STATE);
+
+    /**
+     * XAppStatusIcon::scroll-event:
+     * @icon: The #XAppStatusIcon
+     * @amount: The amount of movement for the scroll event
+     * @direction: the #XAppScrollDirection of the scroll event
+     * @time: The time supplied by the event, or 0
+     *
+     * Gets emitted when the user uses the mouse scroll wheel over the status icon.
+     * For the most part, amounts will always be 1, unless an applet supports smooth
+     * scrolling.  Generally the direction value is most important.
+     */
+    signals [SCROLL] =
+        g_signal_new ("scroll-event",
+                      XAPP_TYPE_STATUS_ICON,
+                      G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
+                      0,
+                      NULL, NULL, NULL,
+                      G_TYPE_NONE, 3, G_TYPE_INT, XAPP_TYPE_SCROLL_DIRECTION, G_TYPE_UINT);
 }
 
 /**
@@ -1327,6 +1341,12 @@ void
 xapp_status_icon_set_name (XAppStatusIcon *icon, const gchar *name)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+
+    if (g_strcmp0 (name, icon->priv->name) == 0)
+    {
+        return;
+    }
+
     g_clear_pointer (&icon->priv->name, g_free);
     icon->priv->name = g_strdup (name);
 
@@ -1337,7 +1357,14 @@ xapp_status_icon_set_name (XAppStatusIcon *icon, const gchar *name)
         xapp_status_icon_interface_set_name (icon->priv->skeleton, name);
     }
 
-    update_fallback_icon (icon);
+    /* Call this directly instead of in the update_fallback_icon() function,
+     * as every time this is called, Gtk re-creates the plug for the icon,
+     * so the tray thinks one icon has disappeared and a new one appeared,
+     * which can cause flicker and undesirable re-ordering of tray items. */
+    if (icon->priv->gtk_status_icon != NULL)
+    {
+        gtk_status_icon_set_name (icon->priv->gtk_status_icon, name);
+    }
 }
 
 /**
@@ -1353,6 +1380,12 @@ void
 xapp_status_icon_set_icon_name (XAppStatusIcon *icon, const gchar *icon_name)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+
+    if (g_strcmp0 (icon_name, icon->priv->icon_name) == 0)
+    {
+        return;
+    }
+
     g_clear_pointer (&icon->priv->icon_name, g_free);
     icon->priv->icon_name = g_strdup (icon_name);
 
@@ -1366,6 +1399,37 @@ xapp_status_icon_set_icon_name (XAppStatusIcon *icon, const gchar *icon_name)
     update_fallback_icon (icon);
 }
 
+/**
+ * xapp_status_icon_get_icon_size:
+ * @icon: a #XAppStatusIcon
+ *
+ * Returns: The desired icon size - usually set by the host based on panel size.
+ * This is not what it's guaranteed to get, and this is really only useful when
+ * receiving absolute icon paths from the client app.
+ *
+ * Since: 1.8
+ */
+gint
+xapp_status_icon_get_icon_size (XAppStatusIcon *icon)
+{
+    g_return_val_if_fail (XAPP_IS_STATUS_ICON (icon), FALLBACK_ICON_SIZE);
+
+    if (icon->priv->skeleton == NULL)
+    {
+        g_debug ("XAppStatusIcon get_icon_size: %d (fallback)", FALLBACK_ICON_SIZE);
+
+        return FALLBACK_ICON_SIZE;
+    }
+
+    gint size;
+
+    size = xapp_status_icon_interface_get_icon_size (icon->priv->skeleton);
+
+    g_debug ("XAppStatusIcon get_icon_size: %d", size);
+
+    return size;
+}
+
 /**
  * xapp_status_icon_set_tooltip_text:
  * @icon: a #XAppStatusIcon
@@ -1379,6 +1443,12 @@ void
 xapp_status_icon_set_tooltip_text (XAppStatusIcon *icon, const gchar *tooltip_text)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+
+    if (g_strcmp0 (tooltip_text, icon->priv->tooltip_text) == 0)
+    {
+        return;
+    }
+
     g_clear_pointer (&icon->priv->tooltip_text, g_free);
     icon->priv->tooltip_text = g_strdup (tooltip_text);
 
@@ -1405,6 +1475,12 @@ void
 xapp_status_icon_set_label (XAppStatusIcon *icon, const gchar *label)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+
+    if (g_strcmp0 (label, icon->priv->label) == 0)
+    {
+        return;
+    }
+
     g_clear_pointer (&icon->priv->label, g_free);
     icon->priv->label = g_strdup (label);
 
@@ -1429,6 +1505,12 @@ void
 xapp_status_icon_set_visible (XAppStatusIcon *icon, const gboolean visible)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+
+    if (visible == icon->priv->visible)
+    {
+        return;
+    }
+
     icon->priv->visible = visible;
 
     g_debug ("XAppStatusIcon set_visible: %s", visible ? "TRUE" : "FALSE");
@@ -1441,6 +1523,63 @@ xapp_status_icon_set_visible (XAppStatusIcon *icon, const gboolean visible)
     update_fallback_icon (icon);
 }
 
+/**
+ * xapp_status_icon_get_visible:
+ * @icon: an #XAppStatusIcon
+ *
+ * Returns whether or not the icon should currently be visible.
+ *
+ * Returns: the current visibility state.
+
+ * Since: 1.8.5
+ */
+gboolean
+xapp_status_icon_get_visible (XAppStatusIcon *icon)
+{
+    g_return_val_if_fail (XAPP_IS_STATUS_ICON (icon), FALSE);
+
+    g_debug ("XAppStatusIcon get_visible: %s", icon->priv->visible ? "TRUE" : "FALSE");
+
+    return icon->priv->visible;
+}
+
+/**
+ * xapp_status_icon_popup_menu:
+ * @icon: an #XAppStatusIcon
+ * @menu: (nullable): A #GtkMenu to display when the primary mouse button is released.
+ * @x: The x anchor position for the menu.
+ * @y: The y anchor position for the menu.
+ * @button: The button used to initiate this action (or 0)
+ * @_time: The event time (or 0)
+ * @panel_position: The #GtkPositionType for the position of the icon.
+ *
+ * Pop up @menu using the positioning arguments. These arguments should be
+ * those provided by a #XAppStatusIcon::button-release-event.
+ *
+ * Since: 1.8.6
+ */
+void
+xapp_status_icon_popup_menu (XAppStatusIcon *icon,
+                             GtkMenu        *menu,
+                             gint            x,
+                             gint            y,
+                             guint           button,
+                             guint           _time,
+                             gint            panel_position)
+
+{
+    g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+    g_return_if_fail (GTK_IS_MENU (menu) || menu == NULL);
+    g_return_if_fail (icon->priv->state != XAPP_STATUS_ICON_STATE_NO_SUPPORT);
+
+    popup_menu (icon,
+                menu,
+                x, y,
+                button,
+                _time,
+                panel_position);
+}
+
 /**
  * xapp_status_icon_set_primary_menu:
  * @icon: an #XAppStatusIcon
@@ -1455,6 +1594,12 @@ xapp_status_icon_set_primary_menu (XAppStatusIcon *icon,
                                    GtkMenu        *menu)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+    g_return_if_fail (GTK_IS_MENU (menu) || menu == NULL);
+
+    if (menu == GTK_MENU (icon->priv->primary_menu))
+    {
+        return;
+    }
 
     g_clear_object (&icon->priv->primary_menu);
 
@@ -1498,9 +1643,15 @@ xapp_status_icon_get_primary_menu (XAppStatusIcon *icon)
  */
 void
 xapp_status_icon_set_secondary_menu (XAppStatusIcon *icon,
-                                   GtkMenu        *menu)
+                                     GtkMenu        *menu)
 {
     g_return_if_fail (XAPP_IS_STATUS_ICON (icon));
+    g_return_if_fail (GTK_IS_MENU (menu) || menu == NULL);
+
+    if (menu == GTK_MENU (icon->priv->secondary_menu))
+    {
+        return;
+    }
 
     g_clear_object (&icon->priv->secondary_menu);
 
